/*
 *      dxemoji.cpp
 *
 *      Copyright 2022 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 * Pango
 * pango-emoji.c: Emoji handling
 *
 * Copyright (C) 2017 Google, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Implementation of pango_emoji_iter is based on Chromium's Ragel-based
 * parser:
 *
 * https://chromium-review.googlesource.com/c/chromium/src/+/1264577
 *
 * The grammar file emoji_presentation_scanner.rl was just modified to
 * adapt the function signature and variables to our usecase.  The
 * grammar itself was NOT modified:
 *
 * https://chromium-review.googlesource.com/c/chromium/src/+/1264577/3/third_party/blink/renderer/platform/fonts/emoji_presentation_scanner.rl
 *
 * The emoji_presentation_scanner.c is generated from .rl file by
 * running ragel on it.
 *
 * The categorization is also based on:
 *
 * https://chromium-review.googlesource.com/c/chromium/src/+/1264577/3/third_party/blink/renderer/platform/fonts/utf16_ragel_iterator.h
 *
 * The iterator next() is based on:
 *
 * https://chromium-review.googlesource.com/c/chromium/src/+/1264577/3/third_party/blink/renderer/platform/fonts/symbols_iterator.cc
 *
 * // Copyright 2015 The Chromium Authors. All rights reserved.
 * // Use of this source code is governed by a BSD-style license that can be
 * // found in the LICENSE file.
 */

#include "dxemoji.h"

FXbool bsearch_interval(FXwchar c, const struct Interval table[], FXuint n)
{
    FXuint lower = 0;
    FXuint upper = n - 1;
    while(lower <= upper)
    {
        int mid = (lower + upper) / 2;
        if (c < table[mid].start)
            upper = mid - 1;
        else if (c > table[mid].end)
            lower = mid + 1;
        else
            return TRUE;
    }
    return FALSE;
}

FXbool dx_Is_Emoji_Base_Character(FXwchar ch)
{
    return dx_Is_Emoji (ch);
}

FXbool dx_Is_Emoji_Extended_Pictographic(FXwchar ch)
{
    return dx_Is_Extended_Pictographic (ch);
}

FXbool dx_Is_Emoji_Emoji_Default(FXwchar ch)
{
  return dx_Is_Emoji_Presentation (ch);
}

FXbool dx_Is_Emoji_Keycap_Base(FXwchar ch)
{
  return (ch >= '0' && ch <= '9') || ch == '#' || ch == '*';
}

FXbool dx_Is_Regional_Indicator(FXwchar ch)
{
  return (ch >= 0x1F1E6 && ch <= 0x1F1FF);
}

unsigned char dx_EmojiSegmentationCategory(FXwchar codepoint)
{
    /* Specific ones first. */
    if (('a' <= codepoint && codepoint <= 'z') || ('A' <= codepoint && codepoint <= 'Z') || codepoint == ' ')
        return kMaxEmojiScannerCategory;

    if ('0' <= codepoint && codepoint <= '9')
        return KEYCAP_BASE;

    switch (codepoint)
    {
    case kCombiningEnclosingKeycapCharacter:
        return COMBINING_ENCLOSING_KEYCAP;
    case kCombiningEnclosingCircleBackslashCharacter:
        return COMBINING_ENCLOSING_CIRCLE_BACKSLASH;
    case kZeroWidthJoinerCharacter:
        return ZWJ;
    case kVariationSelector15Character:
        return VS15;
    case kVariationSelector16Character:
        return VS16;
    case 0x1F3F4:
        return TAG_BASE;
    case 0xE007F:
        return TAG_TERM;
    default: ;
    }

    if((0xE0030 <= codepoint && codepoint <= 0xE0039) || (0xE0061 <= codepoint && codepoint <= 0xE007A))
        return TAG_SEQUENCE;

    if (dx_Is_Emoji_Modifier_Base (codepoint))
        return EMOJI_MODIFIER_BASE;
    if (dx_Is_Emoji_Modifier (codepoint))
        return EMOJI_MODIFIER;
    if (dx_Is_Regional_Indicator (codepoint))
        return REGIONAL_INDICATOR;
    if (dx_Is_Emoji_Keycap_Base (codepoint))
        return KEYCAP_BASE;
    if (dx_Is_Emoji_Emoji_Default (codepoint))
        return EMOJI_EMOJI_PRESENTATION;
    if (dx_Is_Emoji (codepoint))
        return EMOJI_TEXT_PRESENTATION;

    /* Ragel state machine will interpret unknown category as "any". */
    return kMaxEmojiScannerCategory;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wswitch-default"
#include "emoji_presentation_scanner.c"
#pragma GCC diagnostic pop

FXint dx_emoji_end(const FXchar *text, FXint n, bool *isemoji)
{
    FXint result;
    FXint nchars = 0;
    for(FXint pos=0; pos<n; pos+=wclen(text+pos))
    {
        nchars++;
    }
    unsigned char *types = NULL;
    FXMALLOC(&types, unsigned char, nchars);
    FXint sp, sc;
    for (sp = sc = 0; sp < n && sc < nchars; sp += wclen(&text[sp]), sc++)
    {
        types[sc] = dx_EmojiSegmentationCategory(wc(&text[sp]));
    }
    result = scan_emoji_presentation(types, types+nchars, isemoji)-types;
    FXFREE(&types);
    return result;
}

FXint dx_emoji_end(const FXString &text, bool *isemoji)
{
    return dx_emoji_end(text.text(),text.length(),isemoji);
}
