/*
 *      emojisdialog.cpp
 *
 *      Copyright 2022 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "emojisdialog.h"
#include "config.h"
#include "i18n.h"
#include "emojis.h"
#include "dxtextfield.h"

#define SIDE_SPACING             4    // Left or right spacing between items
#define BIG_LINE_SPACING         6    // Line spacing in big icon mode

FXDEFMAP(EmojisList) EmojisListMap[]={
    FXMAPFUNC(SEL_QUERY_TIP,    0,      EmojisList::onQueryTip)
};

FXIMPLEMENT(EmojisList,FXIconList,EmojisListMap,ARRAYNUMBER(EmojisListMap))

EmojisList::EmojisList()
{}

EmojisList::EmojisList(FXComposite* p, FXObject* tgt, FXSelector sel, FXuint opts, FXint x, FXint y, FXint w, FXint h)
    : FXIconList(p, tgt, sel, opts, x, y, w, h)
{}

long EmojisList::onQueryTip(FXObject *sender, FXSelector sel, void *ptr)
{
    if(FXWindow::onQueryTip(sender,sel,ptr)) return 1;
    if((flags&FLAG_TIP) && (0<=cursor))
    {
        FXString string = static_cast<EmojiItem*>(items[cursor])->name();
        sender->handle(this,FXSEL(SEL_COMMAND,ID_SETSTRINGVALUE),(void*)&string);
        return 1;
    }
    return 0;
}

void EmojisList::sortItems()
{
    EmojiItem *v,*c=0;
    FXbool exch=FALSE;
    FXint i,j,h;
    if(0<=current)
    {
        c=static_cast<EmojiItem*>(items[current]);
    }
    for(h=1; h<=items.no()/9; h=3*h+1)
        ;
    for(; h>0; h/=3)
    {
        for(i=h+1;i<=items.no();i++)
        {
            v=static_cast<EmojiItem*>(items[i-1]);
            j=i;
            while(j>h && static_cast<EmojiItem*>(items[j-h-1])->order()>v->order())
            {
                items[j-1]=items[j-h-1];
                exch=TRUE;
                j-=h;
            }
            items[j-1]=v;
        }
    }
    if(0<=current)
    {
        for(i=0; i<items.no(); i++)
        {
            if(items[i]==c){ current=i; break; }
        }
    }
    if(exch) recalc();
}

FXDEFMAP(EmojisDialog) EmojisDialogMap[] = {
    FXMAPFUNC(SEL_CLOSE,            0,                                                  EmojisDialog::onCancel),
    FXMAPFUNC(SEL_COMMAND,          EmojisDialog_CLOSE,                                 EmojisDialog::onClose),
    FXMAPFUNC(SEL_KEYPRESS,         0,                                                  EmojisDialog::onKeyPress),
    FXMAPFUNCS(SEL_DOUBLECLICKED,   EmojisDialog_SMILEYS,EmojisDialog_FILTERED,         EmojisDialog::onTableDoubleClick),
    FXMAPFUNCS(SEL_COMMAND,         EmojisDialog_SMILEYS,EmojisDialog_FLAGS,            EmojisDialog::onToolbar),
    FXMAPFUNC(SEL_CHANGED,          EmojisDialog_SEARCH,                                EmojisDialog::onSearch),
    FXMAPFUNC(SEL_FOCUSIN,          EmojisDialog_SEARCH,                                EmojisDialog::onSearchFocusin)
};

FXIMPLEMENT(EmojisDialog, FXDialogBox, EmojisDialogMap, ARRAYNUMBER(EmojisDialogMap))

EmojisDialog::EmojisDialog(FXApp *app)
        : FXDialogBox(app, _("Select emoji"), DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE, 0,0,0,0, 0,0,0,0, 0,0), m_target(NULL)
{
    m_contents = new FXVerticalFrame(this, LAYOUT_FILL_X|LAYOUT_FILL_Y);

    m_searchframe = new FXHorizontalFrame(m_contents, LAYOUT_FILL_X);
    new FXLabel(m_searchframe, _("Search"));
    m_search = new FXTextField(m_searchframe, 25, this, EmojisDialog_SEARCH, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X);

    m_toolbar = new FXToolBar(m_contents, LAYOUT_DOCK_NEXT|LAYOUT_SIDE_TOP|LAYOUT_FILL_X|FRAME_RAISED);

    m_switcher = new FXSwitcher(m_contents, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK);
    m_smileys = new EmojisList(m_switcher, this, EmojisDialog_SMILEYS);
    m_people = new EmojisList(m_switcher, this, EmojisDialog_PEOPLE);
    m_animals = new EmojisList(m_switcher, this, EmojisDialog_ANIMALS);
    m_food = new EmojisList(m_switcher, this, EmojisDialog_FOOD);
    m_travel = new EmojisList(m_switcher, this, EmojisDialog_TRAVEL);
    m_activities = new EmojisList(m_switcher, this, EmojisDialog_ACTIVITIES);
    m_objects = new EmojisList(m_switcher, this, EmojisDialog_OBJECTS);
    m_symbols = new EmojisList(m_switcher, this, EmojisDialog_SYMBOLS);
    m_flags = new EmojisList(m_switcher, this, EmojisDialog_FLAGS);
    m_filtered = new EmojisList(m_switcher, this, EmojisDialog_FILTERED);

    m_buttonframe = new FXHorizontalFrame(m_contents, LAYOUT_FILL_X);
    m_buttonClose = new dxEXButton(m_buttonframe, _("&Close"), NULL, this, EmojisDialog_CLOSE, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT, 0,0,0,0, 10,10,2,2);

    fillEmojisList();
}

EmojisDialog::EmojisDialog(FXWindow *window)
    : FXDialogBox(window, _("Select emoji"), DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE, 0,0,0,0, 0,0,0,0, 0,0), m_target(window)
{
    m_contents = new FXVerticalFrame(this, LAYOUT_FILL_X|LAYOUT_FILL_Y);

    m_searchframe = new FXHorizontalFrame(m_contents, LAYOUT_FILL_X);
    new FXLabel(m_searchframe, _("Search"));
    m_search = new FXTextField(m_searchframe, 25, this, EmojisDialog_SEARCH, FRAME_THICK|FRAME_SUNKEN|LAYOUT_FILL_X);

    m_toolbar = new FXToolBar(m_contents, LAYOUT_DOCK_NEXT|LAYOUT_SIDE_TOP|LAYOUT_FILL_X|FRAME_RAISED);

    m_switcher = new FXSwitcher(m_contents, LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_THICK);
    m_smileys = new EmojisList(m_switcher, this, EmojisDialog_SMILEYS);
    m_people = new EmojisList(m_switcher, this, EmojisDialog_PEOPLE);
    m_animals = new EmojisList(m_switcher, this, EmojisDialog_ANIMALS);
    m_food = new EmojisList(m_switcher, this, EmojisDialog_FOOD);
    m_travel = new EmojisList(m_switcher, this, EmojisDialog_TRAVEL);
    m_activities = new EmojisList(m_switcher, this, EmojisDialog_ACTIVITIES);
    m_objects = new EmojisList(m_switcher, this, EmojisDialog_OBJECTS);
    m_symbols = new EmojisList(m_switcher, this, EmojisDialog_SYMBOLS);
    m_flags = new EmojisList(m_switcher, this, EmojisDialog_FLAGS);
    m_filtered = new EmojisList(m_switcher, this, EmojisDialog_FILTERED);

    m_buttonframe = new FXHorizontalFrame(m_contents, LAYOUT_FILL_X);
    m_buttonClose = new dxEXButton(m_buttonframe, _("&Close"), NULL, this, EmojisDialog_CLOSE, BUTTON_INITIAL|BUTTON_DEFAULT|FRAME_RAISED|FRAME_THICK|LAYOUT_RIGHT, 0,0,0,0, 10,10,2,2);

    fillEmojisList();
}

EmojisDialog::~EmojisDialog()
{
}

void EmojisDialog::create()
{
    FXDialogBox::create();
    setWidth(700);
    setHeight(350);
}

long EmojisDialog::onCancel(FXObject*, FXSelector, void*)
{
    getApp()->stopModal(this,FALSE);
    hide();
    return 1;
}

long EmojisDialog::onClose(FXObject*, FXSelector, void*)
{
    getApp()->stopModal(this,TRUE);
    hide();
    return 1;
}

long EmojisDialog::onKeyPress(FXObject *sender, FXSelector sel, void *ptr)
{
    if(FXTopWindow::onKeyPress(sender,sel,ptr)) return 1;
    if(((FXEvent*)ptr)->code == KEY_Escape)
    {
        handle(this,FXSEL(SEL_COMMAND,ID_CANCEL),NULL);
        return 1;
    }
    return 0;
}

long EmojisDialog::onSearch(FXObject*, FXSelector, void*)
{
    fillFiltered();
    return 1;
}

long EmojisDialog::onSearchFocusin(FXObject *, FXSelector, void *)
{
    m_filtered->show();
    m_smileys->hide();
    m_people->hide();
    m_animals->hide();
    m_food->hide();
    m_travel->hide();
    m_activities->hide();
    m_objects->hide();
    m_symbols->hide();
    m_flags->hide();
    return 1;
}

void EmojisDialog::fillEmojisList()
{
    FXIcon *smileys = NULL;
    FXIcon *people = NULL;
    FXIcon *animals = NULL;
    FXIcon *food = NULL;
    FXIcon *travel = NULL;
    FXIcon *activities = NULL;
    FXIcon *objects = NULL;
    FXIcon *symbols = NULL;
    FXIcon *flags = NULL;
    for(const auto &it : emojis)
    {
        if(it.second.category == "Smileys & Emotion" && it.second.emoji_order != -1)
        {
            m_smileys->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 2) smileys = it.second.icon;
        }
        if(it.second.category == "People & Body" && it.second.emoji_order != -1)
        {
            m_people->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 227) people = it.second.icon;
        }
        if(it.second.category == "Animals & Nature" && it.second.emoji_order != -1)
        {
            m_animals->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 533) animals = it.second.icon;
        }
        if(it.second.category == "Food & Drink" && it.second.emoji_order != -1)
        {
            m_food->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 721) food = it.second.icon;
        }
        if(it.second.category == "Travel & Places" && it.second.emoji_order != -1)
        {
            m_travel->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 930) travel = it.second.icon;
        }
        if(it.second.category == "Activities" && it.second.emoji_order != -1)
        {
            m_activities->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 1050) activities = it.second.icon;
        }
        if(it.second.category == "Objects" && it.second.emoji_order != -1)
        {
            m_objects->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 1213) objects = it.second.icon;
        }
        if(it.second.category == "Symbols" && it.second.emoji_order != -1)
        {
            m_symbols->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 1479) symbols = it.second.icon;
        }
        if(it.second.category == "Flags" && it.second.emoji_order != -1)
        {
            m_flags->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
            if(it.second.emoji_order == 1650) flags = it.second.icon;
        }
    }
    m_smileys->sortItems();
    m_people->sortItems();
    m_animals->sortItems();
    m_food->sortItems();
    m_travel->sortItems();
    m_activities->sortItems();
    m_objects->sortItems();
    m_symbols->sortItems();
    m_flags->sortItems();
    if(!smileys && m_smileys->getNumItems()) smileys = m_smileys->getItemBigIcon(0);
    if(!people && m_people->getNumItems()) people = m_people->getItemBigIcon(0);
    if(!animals && m_animals->getNumItems()) animals = m_animals->getItemBigIcon(0);
    if(!food && m_food->getNumItems()) food = m_food->getItemBigIcon(0);
    if(!travel && m_travel->getNumItems()) travel = m_travel->getItemBigIcon(0);
    if(!activities && m_activities->getNumItems()) activities = m_activities->getItemBigIcon(0);
    if(!objects && m_objects->getNumItems()) objects = m_objects->getItemBigIcon(0);
    if(!symbols && m_symbols->getNumItems()) symbols = m_symbols->getItemBigIcon(0);
    if(!flags && m_flags->getNumItems()) flags = m_flags->getItemBigIcon(0);
    new FXButton(m_toolbar, FXString("\t")+_("Smileys & Emotion"), smileys, this, EmojisDialog_SMILEYS, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("People & Body"), people, this, EmojisDialog_PEOPLE, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Animals & Nature"), animals, this, EmojisDialog_ANIMALS, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Food & Drink"), food, this, EmojisDialog_FOOD, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Travel & Places"), travel, this, EmojisDialog_TRAVEL, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Activities"), activities, this, EmojisDialog_ACTIVITIES, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Objects"), objects, this, EmojisDialog_OBJECTS, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Symbols"), symbols, this, EmojisDialog_SYMBOLS, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
    new FXButton(m_toolbar, FXString("\t")+_("Flags"), flags, this, EmojisDialog_FLAGS, ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT, 0, 0, 55, 55);
}

void EmojisDialog::fillFiltered()
{
    m_filtered->clearItems();
    if(!m_search->getText().empty())
    {
        FXString search = m_search->getText().lower();
        for(const auto &it : emojis)
        {
            if(it.second.emoji_order != -1)
            {
                if(m_search->getText()[0] == ':')
                {
                    if(it.second.shortname.find(search) != -1) m_filtered->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
                }
                else
                {
                    if(it.second.lowername.find(search) != -1) m_filtered->appendItem(new EmojiItem(it.first.c_str(),it.second.icon,it.second.name,it.second.shortname,it.second.emoji_order));
                }
            }
        }
        m_filtered->sortItems();
        m_switcher->setCurrent(9);
    }
    else
    {
        m_switcher->setCurrent(0);
    }
}

const FXString &EmojisDialog::unicode() const
{
    return m_unicode;
}

void EmojisDialog::setTarget(FXWindow *w)
{
    m_target = w;
}

long EmojisDialog::onTableDoubleClick(FXObject*, FXSelector sel, void *)
{
    switch(FXSELID(sel))
    {
    case EmojisDialog_SMILEYS:
    {
        m_unicode = m_smileys->getItemText(m_smileys->getCurrentItem());
    }break;
    case EmojisDialog_PEOPLE:
    {
        m_unicode = m_people->getItemText(m_people->getCurrentItem());
    }break;
    case EmojisDialog_ANIMALS:
    {
        m_unicode = m_animals->getItemText(m_animals->getCurrentItem());
    }break;
    case EmojisDialog_FOOD:
    {
        m_unicode = m_food->getItemText(m_food->getCurrentItem());
    }break;
    case EmojisDialog_TRAVEL:
    {
        m_unicode = m_travel->getItemText(m_travel->getCurrentItem());
    }break;
    case EmojisDialog_ACTIVITIES:
    {
        m_unicode = m_activities->getItemText(m_activities->getCurrentItem());
    }break;
    case EmojisDialog_OBJECTS:
    {
        m_unicode = m_objects->getItemText(m_objects->getCurrentItem());
    }break;
    case EmojisDialog_SYMBOLS:
    {
        m_unicode = m_symbols->getItemText(m_symbols->getCurrentItem());
    }break;
    case EmojisDialog_FLAGS:
    {
        m_unicode = m_flags->getItemText(m_flags->getCurrentItem());
    }break;
    case EmojisDialog_FILTERED:
    {
        m_unicode = m_filtered->getItemText(m_filtered->getCurrentItem());
    }break;
    default: break;
    }
    if(m_target) m_target->handle(this,FXSEL(SEL_COMMAND,dxTextField::ID_INSERT_STRING),(void*)m_unicode.text());
    return 1;
}

long EmojisDialog::onToolbar(FXObject *, FXSelector sel, void *)
{
    switch(FXSELID(sel))
    {
    case EmojisDialog_SMILEYS:
    {
        m_switcher->setCurrent(0);
    }break;
    case EmojisDialog_PEOPLE:
    {
        m_switcher->setCurrent(1);
    }break;
    case EmojisDialog_ANIMALS:
    {
        m_switcher->setCurrent(2);
    }break;
    case EmojisDialog_FOOD:
    {
        m_switcher->setCurrent(3);
    }break;
    case EmojisDialog_TRAVEL:
    {
        m_switcher->setCurrent(4);
    }break;
    case EmojisDialog_ACTIVITIES:
    {
        m_switcher->setCurrent(5);
    }break;
    case EmojisDialog_OBJECTS:
    {
        m_switcher->setCurrent(6);
    }break;
    case EmojisDialog_SYMBOLS:
    {
        m_switcher->setCurrent(7);
    }break;
    case EmojisDialog_FLAGS:
    {
        m_switcher->setCurrent(8);
    }break;
    default: break;
    }
    return 1;
}

FXIMPLEMENT(EmojiItem,FXIconItem,NULL,0)

EmojiItem::EmojiItem(const FXString &unicode, FXIcon *bi, const FXString &name, const FXString &shortname, FXint order)
    : FXIconItem(unicode, bi), m_name(name), m_shortname(shortname), m_order(order)
{
}

FXint EmojiItem::order() const
{
    return m_order;
}

const FXString &EmojiItem::shortname() const
{
    return m_shortname;
}

const FXString &EmojiItem::name() const
{
    return m_name;
}

void EmojiItem::draw(const FXIconList *list, FXDC &dc, FXint x, FXint y, FXint w, FXint h) const
{
    drawBigIcon(list,dc,x,y,w,h);
}

FXint EmojiItem::hitItem(const FXIconList *list, FXint rx, FXint ry, FXint rw, FXint rh) const
{
    FXint iw=0,tw=0,ih=0,th=0,ss=0,ix,iy,w,h;
    w=list->getItemSpace();
    h=list->getItemHeight();
    if(bigIcon)
    {
        iw=bigIcon->getWidth();
        ih=bigIcon->getHeight();
    }
    iy=BIG_LINE_SPACING/2+(h-th-BIG_LINE_SPACING-ss-ih)/2;
    ix=(w-iw)/2;

    if(ix<=rx+rw && iy<=ry+rh && rx<ix+iw && ry<iy+ih) return 1;
    return 0;
}

void EmojiItem::drawBigIcon(const FXIconList *list, FXDC &dc, FXint x, FXint y, FXint w, FXint h) const
{
    FXint xi,yi;
    FXint iw=0,ih=0,th=0,ss=0;
    if(bigIcon)
    {
        iw=bigIcon->getWidth();
        ih=bigIcon->getHeight();
        xi=x+(w-iw)/2;
        yi=y+BIG_LINE_SPACING/2+(h-th-BIG_LINE_SPACING-ss-ih)/2;
        if(isSelected())
        {
            dc.drawIconShaded(bigIcon,xi,yi);
        }
        else
        {
            dc.drawIcon(bigIcon,xi,yi);
        }
    }
}

EmojiItem::~EmojiItem()
{
}

FXint EmojiItem::getWidth(const FXIconList */*list*/) const
{
    return SIDE_SPACING+(bigIcon?bigIcon->getWidth():0);
}

FXint EmojiItem::getHeight(const FXIconList */*list*/) const
{
    return BIG_LINE_SPACING+(bigIcon?bigIcon->getHeight():0);
}
