/*
 *      emojisdialog.cpp
 *
 *      Copyright 2022 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "emojisdialog.h"
#include "config.h"
#include "emojis.h"
#include "i18n.h"
#include "dxutils.h"

EmojisDialog::EmojisDialog(QWidget *parent) : QDialog(parent), m_smileysModel("Smileys & Emotion", this), m_peopleModel("People & Body", this), m_animalsModel("Animals & Nature", this),
    m_foodModel("Food & Drink", this), m_travelModel("Travel & Places", this), m_activitiesModel("Activities", this), m_objectsModel("Objects", this), m_symbolsModel("Symbols", this),
    m_flagsModel("Flags", this), m_filteredModel("", this)
{
    m_model = new EmojiItemModel(this);
    fillEmojisList();

    QGridLayout *matrix = new QGridLayout;
    matrix->addWidget(new QLabel(_("Search")), 0, 0);
    m_search = new QLineEdit;
    connect(m_search, SIGNAL(textChanged(QString)), this, SLOT(slotSearch()));
    matrix->addWidget(m_search, 0, 1);
    QAction *smileysAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F603.png"), _("Smileys && Emotion"), this);
    connect(smileysAction, SIGNAL(triggered()), this, SLOT(slotSmileys()));
    QAction *peopleAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F9D1.png"), _("People && Body"), this);
    connect(peopleAction, SIGNAL(triggered()), this, SLOT(slotPeople()));
    QAction *animalsAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F436.png"), _("Animals && Nature"), this);
    connect(animalsAction, SIGNAL(triggered()), this, SLOT(slotAnimals()));
    QAction *foodAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F354.png"), _("Food && Drink"), this);
    connect(foodAction, SIGNAL(triggered()), this, SLOT(slotFood()));
    QAction *travelAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/2708.png"), _("Travel && Places"), this);
    connect(travelAction, SIGNAL(triggered()), this, SLOT(slotTravel()));
    QAction *activitiesAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/26BD.png"), _("Activities"), this);
    connect(activitiesAction, SIGNAL(triggered()), this, SLOT(slotActivities()));
    QAction *objectsAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F4A1.png"), _("Objects"), this);
    connect(objectsAction, SIGNAL(triggered()), this, SLOT(slotObjects()));
    QAction *symbolsAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/267B.png"), _("Symbols"), this);
    connect(symbolsAction, SIGNAL(triggered()), this, SLOT(slotSymbols()));
    QAction *flagsAction = new QAction(QIcon(dxutils::getDataPath()+"/icons/emoji/1F1E8-1F1FF.png"), _("Flags"), this);
    connect(flagsAction, SIGNAL(triggered()), this, SLOT(slotFlags()));
    QToolBar *bar = new QToolBar("Emojis", this);
    bar->setStyleSheet("QToolBar { border: 0px }");
    bar->setMovable(false);
    bar->addAction(smileysAction);
    bar->addAction(peopleAction);
    bar->addAction(animalsAction);
    bar->addAction(foodAction);
    bar->addAction(travelAction);
    bar->addAction(activitiesAction);
    bar->addAction(objectsAction);
    bar->addAction(symbolsAction);
    bar->addAction(flagsAction);
    EmojiItemDelegate *delegate = new EmojiItemDelegate();
    m_switch = new QStackedWidget;
    m_smileys = new QListView;
    m_smileys->setItemDelegate(delegate);
    m_smileys->setViewMode(QListWidget::IconMode);
    m_smileys->setResizeMode(QListView::Adjust);
    m_smileysModel.setSourceModel(m_model);
    m_smileys->setModel(&m_smileysModel);
    m_smileys->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_smileys, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_people = new QListView;
    m_people->setItemDelegate(delegate);
    m_people->setViewMode(QListWidget::IconMode);
    m_people->setResizeMode(QListView::Adjust);
    m_peopleModel.setSourceModel(m_model);
    m_people->setModel(&m_peopleModel);
    m_people->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_people, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_animals = new QListView;
    m_animals->setItemDelegate(delegate);
    m_animals->setViewMode(QListWidget::IconMode);
    m_animals->setResizeMode(QListView::Adjust);
    m_animalsModel.setSourceModel(m_model);
    m_animals->setModel(&m_animalsModel);
    m_animals->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_animals, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_food = new QListView;
    m_food->setItemDelegate(delegate);
    m_food->setViewMode(QListWidget::IconMode);
    m_food->setResizeMode(QListView::Adjust);
    m_foodModel.setSourceModel(m_model);
    m_food->setModel(&m_foodModel);
    m_food->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_food, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_travel = new QListView;
    m_travel->setItemDelegate(delegate);
    m_travel->setViewMode(QListWidget::IconMode);
    m_travel->setResizeMode(QListView::Adjust);
    m_travelModel.setSourceModel(m_model);
    m_travel->setModel(&m_travelModel);
    m_travel->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_travel, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_activities = new QListView;
    m_activities->setItemDelegate(delegate);
    m_activities->setViewMode(QListWidget::IconMode);
    m_activities->setResizeMode(QListView::Adjust);
    m_activitiesModel.setSourceModel(m_model);
    m_activities->setModel(&m_activitiesModel);
    m_activities->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_activities, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_objects = new QListView;
    m_objects->setItemDelegate(delegate);
    m_objects->setViewMode(QListWidget::IconMode);
    m_objects->setResizeMode(QListView::Adjust);
    m_objectsModel.setSourceModel(m_model);
    m_objects->setModel(&m_objectsModel);
    m_objects->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_objects, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_symbols = new QListView;
    m_symbols->setItemDelegate(delegate);
    m_symbols->setViewMode(QListWidget::IconMode);
    m_symbols->setResizeMode(QListView::Adjust);
    m_symbolsModel.setSourceModel(m_model);
    m_symbols->setModel(&m_symbolsModel);
    m_symbols->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_symbols, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_flags = new QListView;
    m_flags->setItemDelegate(delegate);
    m_flags->setViewMode(QListWidget::IconMode);
    m_flags->setResizeMode(QListView::Adjust);
    m_flagsModel.setSourceModel(m_model);
    m_flags->setModel(&m_flagsModel);
    m_flags->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_flags, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    m_filtered = new QListView;
    m_filtered->setItemDelegate(delegate);
    m_filtered->setViewMode(QListWidget::IconMode);
    m_filtered->setResizeMode(QListView::Adjust);
    m_filteredModel.setSourceModel(m_model);
    m_filtered->setModel(&m_filteredModel);
    m_filtered->setEditTriggers(QAbstractItemView::NoEditTriggers);
    connect(m_filtered, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(slotDoubleClick(QModelIndex)));
    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Close);
    connect(buttons, SIGNAL(rejected()), this, SLOT(reject()));
    QVBoxLayout *mainLayout = new QVBoxLayout;
    mainLayout->addLayout(matrix);
    mainLayout->addWidget(bar);
    m_switch->addWidget(m_smileys);
    m_switch->addWidget(m_people);
    m_switch->addWidget(m_animals);
    m_switch->addWidget(m_food);
    m_switch->addWidget(m_travel);
    m_switch->addWidget(m_activities);
    m_switch->addWidget(m_objects);
    m_switch->addWidget(m_symbols);
    m_switch->addWidget(m_flags);
    m_switch->addWidget(m_filtered);
    mainLayout->addWidget(m_switch);
    mainLayout->addWidget(buttons);
    setLayout(mainLayout);
    setWindowTitle(_("Select emoji"));

    m_smileysModel.sort(0);
    m_peopleModel.sort(0);
    m_animalsModel.sort(0);
    m_foodModel.sort(0);
    m_travelModel.sort(0);
    m_activitiesModel.sort(0);
    m_objectsModel.sort(0);
    m_symbolsModel.sort(0);
    m_flagsModel.sort(0);
    m_filteredModel.sort(0);

    resize(520,350);
}

void EmojisDialog::fillEmojisList()
{
    for(const auto &it : qAsConst(emojis))
    {
        if(it.emoji_order != -1) m_model->appendRow(new EmojiItem(it.unicode,it.filename,it.name,it.shortname,it.category,it.emoji_order));
    }
}

void EmojisDialog::accept()
{
    QDialog::accept();
}

void EmojisDialog::slotDoubleClick(QModelIndex index)
{
    if(index.isValid())
    {
        emit unicodeClicked(index.data(EmojiItem::EUNICODE).toString());
    }
}

void EmojisDialog::slotSmileys()
{
    m_switch->setCurrentIndex(0);
}

void EmojisDialog::slotPeople()
{
    m_switch->setCurrentIndex(1);
}

void EmojisDialog::slotAnimals()
{
    m_switch->setCurrentIndex(2);
}

void EmojisDialog::slotFood()
{
    m_switch->setCurrentIndex(3);
}

void EmojisDialog::slotTravel()
{
    m_switch->setCurrentIndex(4);
}

void EmojisDialog::slotActivities()
{
    m_switch->setCurrentIndex(5);
}

void EmojisDialog::slotObjects()
{
    m_switch->setCurrentIndex(6);
}

void EmojisDialog::slotSymbols()
{
    m_switch->setCurrentIndex(7);
}

void EmojisDialog::slotFlags()
{
    m_switch->setCurrentIndex(8);
}

void EmojisDialog::slotSearch()
{
    if(m_search->text().isEmpty())
    {
        m_switch->setCurrentIndex(0);
    }
    else
    {
        m_filteredModel.setSearch(m_search->text());
        m_switch->setCurrentIndex(9);
    }
}

EmojiItem::EmojiItem(const QString &unicode, const QString &filename, const QString &name, const QString &shortname, const QString &category, int order)
{
    setText("");
    setData(unicode, EUNICODE);
    setIcon(QIcon(filename));
    setData(name, ENAME);
    setData(shortname, ESHORTNAME);
    setData(category, ECATEGORY);
    setData(order, EORDER);
    setData(filename, EFILENAME);
    setToolTip(name);
}

EmojiItemDelegate::EmojiItemDelegate()
{
}

QSize EmojiItemDelegate::sizeHint(const QStyleOptionViewItem &/*option*/, const QModelIndex &/*index*/) const
{
    return QSize(52,52);
}

void EmojiItemDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    painter->save();
    painter->setRenderHint(QPainter::TextAntialiasing);
    QVariant value = index.data(Qt::DecorationRole);
    QColor selectedBg(60, 140, 222);
    QPen penDivision;
    if(option.state & QStyle::State_Selected)
    {
        painter->fillRect(option.rect, selectedBg);
        penDivision.setColor(selectedBg);
    }
#if (QT_VERSION >= QT_VERSION_CHECK(6, 0, 0))
    if(value.metaType().id() == QMetaType::QIcon)
#else
    if(value.type() == QVariant::Icon)
#endif
    {
        QIcon icon = qvariant_cast<QIcon>(value);
        QRect rect(option.rect.x()+2,option.rect.y()+2,48,48);
        icon.paint(painter,rect);
    }
    painter->restore();
}

EmojiItemSortModel::EmojiItemSortModel(const QString &category, QObject *parent)
    : QSortFilterProxyModel(parent), m_search(""), m_category(category)
{
    setDynamicSortFilter(true);
    setFilterRole(EmojiItem::EORDER);
}

void EmojiItemSortModel::setSearch(const QString &newSearch)
{
    m_search = newSearch;
    invalidateFilter();
}

bool EmojiItemSortModel::lessThan(const QModelIndex &left, const QModelIndex &right) const
{
    return left.data(EmojiItem::EORDER).toInt() < right.data(EmojiItem::EORDER).toInt();
}

bool EmojiItemSortModel::filterAcceptsRow(int source_row, const QModelIndex &source_parent) const
{
    if(m_search.isEmpty())
    {
        if(!m_category.isEmpty())
        {
            QModelIndex index = sourceModel()->index(source_row, 0, source_parent);
            return m_category == sourceModel()->data(index, EmojiItem::ECATEGORY).toString();
        }
        else
        {
            return true;
        }
    }
    QModelIndex index = sourceModel()->index(source_row, 0, source_parent);
    QString lower = sourceModel()->data(index, EmojiItem::ENAME).toString().toLower();
    QString shortname = sourceModel()->data(index, EmojiItem::ESHORTNAME).toString();
    if(m_search[0] == ':')
    {
        return shortname.contains(m_search);
    }
    else
    {
        return lower.contains(m_search);
    }
}

EmojiItemModel::EmojiItemModel(QObject *parent)
    : QStandardItemModel(parent)
{
}

QString EmojiItemModel::getUnicode(int row) const
{
    for(int r = 0; r < rowCount(); ++r)
    {
        EmojiItem *i = static_cast<EmojiItem*>(item(r));
        if(row == r)
            return i->data(EmojiItem::EUNICODE).toString();
    }
    return "";
}
